import os
import random
import sys
import time

import pygame
from pygame.sprite import Sprite

import maze
import namer

scribe = namer.Namer()
scribe.learnfile('namer/base.txt')

adjectives = [s.strip() for s in open('adjs.txt') if len(s.strip()) < 10]

deltas = [(-1, 0), (0, 1), (1, 0), (0, -1)]

def tint_image(img, tinter=None, limit=20):
    if not tinter: return
    for y in range(0, limit):
        for x in range(0, limit):
            c = img.get_at((x, y))
            c.r = tinter(c.r)
            c.g = tinter(c.g)
            c.b = tinter(c.b)
            img.set_at((x, y), c)

def c_to_list(c):
    return [c.r, c.g, c.b]
def c_from_list(ls, a):
    return pygame.Color(ls[0], ls[1], ls[2], a)

def swap_image(img, locs=[0, 1, 2], limit=20):
    for y in range(0, limit):
        for x in range(0, limit):
            c = img.get_at((x, y))
            ls1 = c_to_list(c)

            ls2 = [0, 0, 0]
            ls2[locs[0]] = ls1[0]
            ls2[locs[1]] = ls1[1]
            ls2[locs[2]] = ls1[2]

            c = c_from_list(ls2, c.a)
            img.set_at((x, y), c)

class Tile(Sprite):
    def __init__(self, screen, image, size=20):
        Sprite.__init__(self)
        self.screen = screen
        self.base_image = image
        self.image = image.copy()
        self.size = size

    @classmethod
    def frompath(cls, screen, path):
        image = pygame.image.load(path).convert_alpha()
        return cls(screen, image)

    @classmethod
    def scalefrompath(cls, screen, path, fps=4, scale=1):
        images = []
        s1 = pygame.image.load(path).convert_alpha()
        s2 = pygame.Surface((20 * scale, 20 * scale), depth=32).convert_alpha()
        for y in range(0, 20):
            for x in range(0, 20):
                c = s1.get_at((x, y))
                y2, x2 = y * scale, x * scale
                for i in range(0, scale):
                    for j in range(0, scale):
                        s2.set_at((x2 + i, y2 + j), c)
        return cls(screen, s2, size=20 * scale)

    def tint(self, tinter=None):
        self.image = self.base_image.copy()
        if tinter is None: return
        tint_image(self.image, tinter)

    def blitme(self, pos):
        self.screen.blit(self.image, pos)

class AnimatedSprite(pygame.sprite.Sprite):
    def __init__(self, screen, images, fps=4, size=20):
        pygame.sprite.Sprite.__init__(self)
        self.base_images = images
        self.images = [img.copy() for img in self.base_images]
        self.image = None

        self.size = size
        self.num = len(images)
        self.screen = screen
 
        self.fps = fps
        self.start = pygame.time.get_ticks()
        self.delay = 1000 / fps
        self.last_update = 0
        self.frame = 0

        self.update(pygame.time.get_ticks())
 
    @classmethod
    def frompaths(cls, screen, paths, fps=4):
        images = [pygame.image.load(path).convert_alpha() for path in paths]
        return cls(screen, images, fps)

    @classmethod
    def scalefrompaths(cls, screen, paths, fps=4, scale=1):
        images = []
        for path in paths:
            s1 = pygame.image.load(path).convert_alpha()
            s2 = pygame.Surface((20 * scale, 20 * scale), depth=32).convert_alpha()
            for y in range(0, 20):
                for x in range(0, 20):
                    c = s1.get_at((x, y))
                    y2, x2 = y * scale, x * scale
                    for i in range(0, scale):
                        for j in range(0, scale):
                            s2.set_at((x2 + i, y2 + j), c)
            images.append(s2)
        return cls(screen, images, fps, size=20 * scale)

    def update(self, t):
        dt = t - self.last_update
        n = dt / self.delay
        r = dt % self.delay

        if n > 0:
            self.frame = (self.frame + n) % self.num
            self.image = self.images[self.frame]
            self.last_update = t - r

    def tint(self, tinter=None):
        self.images = [img.copy() for img in self.base_images]
        if tinter is None: return
        for img in self.images: tint_image(img, tinter, limit=self.size)
        self.image = self.images[self.frame]

    def swaprgb(self, locs):
        self.images = [img.copy() for img in self.base_images]
        for img in self.images: swap_image(img, locs, limit=self.size)
        self.image = self.images[self.frame]

    def blitme(self, pos, inverse=False):
        if inverse:
            img = self.image.copy()
            tint_image(img, lambda x: x ^ 255)
            self.screen.blit(img, pos)
        else:
            self.screen.blit(self.image, pos)

class EntitySprite(AnimatedSprite):
    def copy_images(self, images, size=20):
        return [img.copy() for img in images]

    def __init__(self, screen, north, east, south, west, fps=4):
        self.facing = 2
        self.base_sprites = [north, east, south, west]
        self.sprites = [self.copy_images(imgs) for imgs in self.base_sprites]
        
        AnimatedSprite.__init__(self, screen, self.sprites[self.facing], fps)

    def copy(self):
        s = [self.copy_images(imgs) for imgs in self.base_sprites]
        return EntitySprite(self.screen, s[0], s[1], s[2], s[3], self.fps)

    @classmethod
    def frompaths(cls, screen, npaths, epaths, spaths, wpaths, fps=4):
        north = [pygame.image.load(path).convert_alpha() for path in npaths]
        east =  [pygame.image.load(path).convert_alpha() for path in epaths]
        south = [pygame.image.load(path).convert_alpha() for path in spaths]
        west =  [pygame.image.load(path).convert_alpha() for path in wpaths]
        return cls(screen, north, east, south, west, fps)

    def turn(self, direction):
        assert direction <= 3
        assert direction >= 0
        self.facing = direction
        self.images = self.sprites[self.facing]
        self.image = self.images[self.frame]

    def swaprgb(self, locs):
        self.sprites = [self.copy_images(imgs) for imgs in self.base_sprites]
        for images in self.sprites:
            for img in images:
                swap_image(img, locs)
        self.base_images = self.sprites[self.facing]
        self.images = self.sprites[self.facing]
        self.image = self.images[self.frame]

    def tint(self, tinter=None):
        self.sprites = [self.copy_images(imgs) for imgs in self.base_sprites]

        if tinter is None: return
        for images in self.sprites:
            for img in images:
                tint_image(img, tinter)
        self.base_images = self.sprites[self.facing]
        self.images = self.sprites[self.facing]
        self.image = self.images[self.frame]

class Entity(object):
    moves = [(-1, 0), (0, 1), (1, 0), (0, -1)]
    speed = 10
    stotal = 100
    sunit = 5

    def __init__(self, sprite, y, x):
        self.sprite = sprite
        self.y = y
        self.x = x
        self.yy = 0
        self.xx = 0

        self.pushed = None
        self.pspeed = 1
        self.pcounter = 0

        self.damaged = 0

    def copy(self):
        return Entity(self.sprite.copy(), self.y, self.x)

    def damage(self, game, direction, duration=40):
        self.damaged = duration
        self.startpush(direction, 4, 40)

    def startpush(self, direction, speed=2, counter=10):
        self.pushed = direction
        self.pspeed = self.stotal
        self.pcounter = counter

    def ispushed(self):
        return self.pushed is not None

    def keeppushing(self, game):
        for i in range(0, self.pspeed):
            self.push(self.pushed, game)
            self.pcounter -= 1
            if self.pcounter < 1:
                self.pushed = None
                break

    def ismoving(self):
        return self.yy != 0 or self.xx != 0

    def keepmoving(self, game):
        self.move(self.sprite.facing, game)

    def moveok(self, direction, game):
        dy, dx = self.moves[direction]
        if dy:
            y2 = self.y + dy
            self.xx = 0
            #if self.xx != 0: return (False, 0, 0)
            if y2 < 0 or y2 >= game.theight: return (False, 0, 0)
            if game.maze.cave[y2][self.x].iswall: return (False, 0, 0)
        else:
            x2 = self.x + dx
            self.yy = 0
            #if self.yy != 0: return (False, 0, 0)
            if x2 < 0 or x2 >= game.twidth: return (False, 0, 0)
            if game.maze.cave[self.y][x2].iswall: return (False, 0, 0)
        return (True, dy, dx)

    def move(self, direction, game):
        if self.sprite.facing != direction:
            self.sprite.turn(direction)
        self.push(direction, game)

    def push(self, direction, game):
        ok, dy, dx = self.moveok(direction, game)
        if not ok: return

        if dy:
            y2 = self.y + dy
            self.yy += dy * self.speed
            if self.yy >= self.stotal:
                self.y += 1
                self.yy = 0
            elif self.yy <= -self.stotal:
                self.y -= 1
                self.yy = 0
        else:
            x2 = self.x + dx
            self.xx += dx * self.speed
            if self.xx >= self.stotal:
                self.x += 1
                self.xx = 0
            elif self.xx <= -self.stotal:
                self.x -= 1
                self.xx = 0

    def getpos(self):
        return (self.y * 20 + self.yy / self.sunit,
                self.x * 20 + self.xx / self.sunit)

    def getbox(self):
        y, x = self.getpos()
        return (y, x, y + 20, x + 20)

class Kind(object):
    path = None
    pinned = False
    rim = 0
    wait = None
    def __init__(self, screen):
        self.small = Tile.frompath(screen, self.path)
        self.big = Tile.scalefrompath(screen, self.path, scale=2)
    def act(self, game, item):
        item.acting = True
        self._act(game, item)
    def _act(self, game, item): raise Exception('implement me')
    def update(self, game, item): raise Exception('implement me')

    def render(self, game, item, y, x):
        y2, x2 = self.getpos(game, item)
        self.small.blitme((x2 + game.xoff, y2 + game.yoff))

    def cancel(self, item): item.acting = False
    def doeffect(self, game, item): pass

    def getpos(self, game, item):
        return game.pentity.getpos()
    def getbox(self, game, item):
        y, x = self.getpos(game, item)
        return (y - self.rim, x - self.rim, y + 20 + self.rim, x + 20 + self.rim)

class Shield(Kind):
    dist = 0
    path = 'shield.png'
    time = 1
    pinned = True
    dist = 1
    wait = pygame.K_SPACE
    def _act(self, game, item):
        game.curr_player().immune = True
    def cancel(self, game, item):
        game.curr_player().immune = False
        item.acting = False
    def update(self, game, item):
        if item.acting:
            return self
        else:
            return None


class Ring(Kind):
    dist = 0
    path = 'ring.png'
    time = 1
    pinned = True
    dist = 1
    def _act(self, game, item):
        game.telesnd.play()
        game.pentity.y = random.randint(0, 4) * 2 + 1
        game.pentity.x = random.randint(0, 4) * 2 + 1
        game.pentity.xx = game.pentity.yy = 0
        game.pentity.damaged = 20
  
    def update(self, game, item):
        return None

class Wand(Kind):
    dist = 0
    path = 'wand.png'
    time = 1
    pinned = True
    dist = 1
    def _act(self, game, item):
        if game.gold < 3:
            return
        game.bighitsnd.play()
        game.gold -= 3
        for monster in game.monsters:
            monster.damage(game, 0, 1000)
  
    def update(self, game, item):
        return None

class Weapon(Kind):
    dmg = 10
    def doeffect(self, game, item):
        iy1, ix1, iy2, ix2 = self.getbox(game, item)

        for monster in game.monsters:
            if monster.isdamaged():
                continue

            my1, mx1, my2, mx2 = monster.entity.getbox()
            if my1 >= iy2 or my2 <= iy1: continue
            if mx1 >= ix2 or mx2 <= ix1: continue

            #print (iy1, ix1, iy2, ix2), (my1, mx1, my2, mx2)
            #print (iy1, ix1, iy2, ix2), (my1, mx1, my2, mx2)

            if my1 == iy1:
                d = 2 + cmp(mx1, ix1)
            elif mx1 == ix1:
                d = 1 + cmp(iy1, my1)
            else:
                d = (monster.entity.sprite.facing + 2) % 4
            monster.damage(game, d, self.dmg)

    def hurt(self, game, item, monster):
        pass

class Cat(Weapon):
    path = 'cat.png'
    time = 20
    halfway = time / 2
    pixelper = 4
    def _act(self, game, item):
        item.duration = self.time
        item.direction = game.pentity.sprite.facing
    def update(self, game, item):
        if not item.acting: return
        item.duration -= 1
        if item.duration == 0:
            item.acting = False
            return None
        return self
    def getpos(self, game, item):
        yd, xd = deltas[item.direction]

        if item.duration > self.halfway:
            d = self.time - item.duration
        else:
            d = item.duration

        by, bx = game.pentity.getpos()
        y2 = by + yd * (d * self.pixelper)
        x2 = bx + xd * (d * self.pixelper)
        return (y2, x2)

class Swing(Weapon):
    dist = 1
    def _act(self, game, item):
        item.duration = self.time
        game.bighitsnd.play()
    def update(self, game, item):
        if not item.acting: return
        item.duration -= 1
        if item.duration == 0:
            item.acting = False
            return None
        return self
    def getpos(self, game, items):
        yd, xd = deltas[game.pentity.sprite.facing]
        by, bx = game.pentity.getpos()
        y2 = by + yd * 20 * self.dist
        x2 = bx + xd * 20 * self.dist
        return (y2, x2)
    def render(self, game, item, y, x):
        yd, xd = deltas[game.pentity.sprite.facing]
        by, bx = game.pentity.getpos()
        y2 = by + yd * 20 * self.dist
        x2 = bx + xd * 20 * self.dist
        self.small.blitme((x2 + game.xoff, y2 + game.yoff))

#class Crossbow(Weapon):
class Crossbow(Swing):
    dist = 1
    path = 'xbow.png'
    bpath = 'bolt.png'
    time = 18
    pixelper = 10
    pinned = True
    def __init__(self, screen):
        Weapon.__init__(self, screen)
        self.bolt = Tile.scalefrompath(screen, self.bpath)
    def _act(self, game, item):
        p = game.curr_player()
        item.duration = self.time
        item.direction = game.pentity.sprite.facing
        item.basepos = game.pentity.getpos()
        game.time = max(game.time - 2, 0)

    def update(self, game, item):
        if not item.acting: return
        item.duration -= 1
        if item.duration == 0:
            item.acting = False
            return None
        return self

    def getpos2(self, game, item):
        yd, xd = deltas[item.direction]
        sofar = (self.time - item.duration) * self.pixelper
    
        y2 = item.basepos[0] + sofar * yd
        x2 = item.basepos[1] + sofar * xd
        return (y2, x2)
    def getbox2(self, game, item):
        y, x = self.getpos2(game, item)
        return (y, x, y + 20, x + 20)

    def render(self, game, item, y, x):
        Weapon.render(self, game, item, y, x)

        y2, x2 = self.getpos2(game, item)
        if y2 < 20 or y2 > 180: return
        if x2 < 20 or x2 > 180: return
        self.bolt.blitme((x2 + game.xoff, y2 + game.yoff))

    def doeffect(self, game, item):
        iy1, ix1, iy2, ix2 = self.getbox2(game, item)

        for monster in game.monsters:
            if monster.isdamaged():
                continue

            my1, mx1, my2, mx2 = monster.entity.getbox()
            if my1 >= iy2 or my2 <= iy1: continue
            if mx1 >= ix2 or mx2 <= ix1: continue

            #print (iy1, ix1, iy2, ix2), (my1, mx1, my2, mx2)
            #print (iy1, ix1, iy2, ix2), (my1, mx1, my2, mx2)

            if my1 == iy1:
                d = 2 + cmp(mx1, ix1)
            elif mx1 == ix1:
                d = 1 + cmp(iy1, my1)
            else:
                d = (monster.entity.sprite.facing + 2) % 4
            monster.damage(game, d, self.dmg)

class Axe(Swing):
    path = 'axe.png'
    time = 18
    pinned = True

class Sword(Swing):
    path = 'sword.png'
    time = 8
    pinned = True

class Item(object):
    isgold = False
    istime = False
    ishealth = False
    def __init__(self, kind):
        self.kind = kind

class TimeItem(Item):
    istime = True

class GoldItem(Item):
    isgold = True

class HealthItem(Item):
    ishealth = True

class Player(object):
    def __init__(self, entity, locs, inventory=[]):
        self.entity = entity
        self.entity.speed = 10
        self.name = scribe.generate().upper()
        self.title = random.choice(adjectives).upper()
        self.locs = locs
        self.hp = 6
        self.maxhp = 6
        self.inventory = (inventory + [None, None, None])[:3]
        self.equipped = 0
        self.immune = False
    def update(self, game):
        p = game.curr_player()
        lvl = game.curr_level()
        if self.entity.damaged:
            self.entity.damaged -= 1
        if self.entity.yy == 0 and self.entity.xx == 0:
            key = (self.entity.y, self.entity.x)
            if key in lvl.items:
                item = lvl.items[key]
                used = True
                if item.isgold:
                    game.gold += 1
                    game.coinsnd.play()
                elif item.istime:
                    game.time += 30
                    game.timesnd.play()
                    game.dirtytime = True
                elif item.ishealth:
                    if p.hp == p.maxhp:
                        used = False
                    else:
                        game.healthsnd.play()
                        p.hp = min(p.maxhp, p.hp + 2)
                else:
                    if self.inventory[0] is None:
                        self.inventory[0] = item
                    elif self.inventory[1] is None:
                        self.inventory[1] = item
                    elif self.inventory[2] is None:
                        self.inventory[2] = item
                    else:
                        used = False
                    if used: game.itemsnd.play()
                    if used and game.trading:
                        op = game.other_player()
                        for i in range(0, 3):
                            if op.inventory[i] == item:
                                op.inventory[i] = None
                if used:
                    del lvl.items[key]
                    game.update_stats()

    def damage(self, game, direction):
        game.hurtsnd.play()
        self.hp -= 1
        if self.hp < 1: return
        self.entity.damage(game, direction)
    def isdead(self):
        return self.hp < 1
    def isimmune(self):
        return self.immune or self.entity.damaged
    def isdamaged(self):
        return self.entity.damaged
    def isinverse(self):
        return self.entity.damaged % 2 == 1

class Brain(object):
    @classmethod
    def doturn(self, game, monster):
        if monster.entity.ismoving():
            monster.entity.keepmoving(game)
            return
        self._doturn(game, monster)

    @classmethod
    def _doturn(self, game, monster):
        raise Exception("implement me")

class RandomMover(Brain):
    @classmethod
    def _doturn(self, game, monster):
        y, x = monster.entity.y, monster.entity.x
        od = monster.entity.sprite.facing
        ds = list(game.maze.open_directions(y, x))
        if len(ds) > 1:
            ds = [d for d in ds if (d - od) % 4 != 2]
        d = random.choice(ds)
        monster.entity.move(d, game)

class Turner(Brain):
    order = None
    @classmethod
    def _doturn(self, game, monster):
        od = monster.entity.sprite.facing
        for i in self.order:
            d = (od + i) % 4
            ok, _, _ = monster.entity.moveok(d, game)
            if ok:
                monster.entity.move(d, game)
                return

class RightTurner(Turner): order = [1, 0, -1, 2]
class LeftTurner(Turner): order = [-1, 0, 1, 2]

class Species(object):
    maxhp = 10
    brains = [RandomMover]
    def __init__(self, sprite):
        self.sprite = sprite

    def create(self, y, x, speed=10):
        brain = random.choice(self.brains)
        e = Entity(self.sprite.copy(), y, x)
        e.speed = speed
        return Monster(e, self, brain)

class Goblin(Species):
    brains = [RandomMover, RandomMover, LeftTurner]

class Monster(object):
    def __init__(self, entity, species, brain):
        self.entity = entity
        self.species = species
        self.hp = species.maxhp
        self.brain = brain
    def update(self, game):
        if self.entity.damaged:
            self.entity.damaged -= 1
    def isdead(self):
        return self.hp < 1
    def damage(self, game, direction, amt=0):
        game.bighitsnd.play()
        self.hp -= amt
        if self.hp < 1: return
        self.entity.damage(game, direction, 40)
    def isdamaged(self):
        return self.entity.damaged

class Zone(object):
    possible = [(0, 1, 2), (0, 2, 1),
                (1, 0, 2), (1, 2, 0),
                (2, 0, 1), (2, 1, 0)]
    def __init__(self, game, levels, forbid=None):
        self.game = game

        self.tint = random.randint(0, 255)

        choices = [t for t in self.possible if t != forbid]
        locs = random.choice(choices)

        if game.zindex == 0:
            i1 = Item(self.game.kinds['SWORD'])
        else:
            i1 = None
        self.player = Player(game.pentity.copy(), locs, [i1, None, None])
        self.levels = levels

class TradeLevel(object):
    def __init__(self):
        self.prev = None
        self.seed = None
        self.next = None

        self.items = {}

class Level(object):
    def place_random_item(self, item):
        x = random.randint(0, 4) * 2 + 1
        y = random.randint(0, 4) * 2 + 1
        if (y, x) != (self.sy, self.sx):
            self.items[(y, x)] = item

    def __init__(self, game, prev, seed, sy, sx):
        if game.zones:
            p = game.curr_player()
            pkinds = set([i.kind for i in p.inventory if i is not None])
        else:
            pkinds = set([game.kinds['SWORD']])

        self.prev = prev
        self.seed = seed
        self.next = random.randint(0, 1 << 24)

        self.tint = random.randint(0, 255)
        self.sy = sy
        self.sx = sx

        self.items = {}

        difficulty = game.get_difficulty()

        ngold = max(1, random.randint(game.zindex, game.zindex * 2))
        for i in range(0, ngold):
            self.place_random_item(GoldItem(None))

        if random.randint(0, 2) == 0:
            self.place_random_item(HealthItem(None))

        if game.zindex > 0 and random.randint(0, 3) == 0:
            self.place_random_item(TimeItem(None))

        if game.lindex > 0 and random.randint(0, 4) == 0:
            kinds = [k for k in game.kinds.values() if k not in pkinds]
            item = Item(random.choice(kinds))
            self.place_random_item(item)

class Game(object):
    twidth = 11
    theight = 11
    bgcolor = (0, 0, 30)

    yoff = 20
    xoff = 20

    def get_difficulty(self):
        return self.zindex * 3 + self.lindex

    def mktext(self, s, color=None):
        if color is None:
            color = pygame.Color(255,255,255,255)
        return self.font.render(s, False, color)

    def __init__(self, screen):
        self.screen = screen

        self.font = pygame.font.Font("emulator.ttf", 18, bold=True)

        self.starttile = Tile.frompath(screen, 'start.png')
        self.walltile = Tile.frompath(screen, 'wall.png')
        self.opentile = Tile.frompath(screen, 'open.png')
        self.exittile = Tile.frompath(screen, 'exit.png')
        self.debugtile = Tile.frompath(screen, 'debug.png')
        self.doortile = Tile.frompath(screen, 'door.png')
        self.selecttile = Tile.frompath(screen, 'select.png')
        self.goldtile = Tile.frompath(screen, 'gold.png')
        self.timetile = Tile.frompath(screen, 'time.png')
        self.healthtile = Tile.frompath(screen, 'health.png')
        self.firetile = Tile.frompath(screen, 'fire.png')

        self.healthsnd = pygame.mixer.Sound("sfx/health.wav")
        self.coinsnd = pygame.mixer.Sound("sfx/coin.wav")
        self.timesnd = pygame.mixer.Sound("sfx/time.wav")
        self.itemsnd = pygame.mixer.Sound("sfx/item.wav")
        self.telesnd = pygame.mixer.Sound("sfx/tele.wav")
        self.hurtsnd = pygame.mixer.Sound("sfx/hurt.wav")
        self.bighitsnd = pygame.mixer.Sound("sfx/bighit.wav")
        self.hihit = pygame.mixer.Sound("sfx/hihit.wav")

        self.kinds = {
            'CAT': Cat(screen),
            'CROSSBOW': Crossbow(screen),
            'SWORD': Sword(screen),
            'AXE': Axe(screen),
            'SHIELD': Shield(screen),
            'RING': Ring(screen),
            'WAND': Wand(screen),
        }

        self.trading = False
        self.trademaze = TradeMaze(self.theight, self.twidth)
        self.tradelevel = TradeLevel()

        self.gold = 0

        self.psprite = EntitySprite.frompaths(screen,
                                              ['melf/melfn1.png', 'melf/melfn2.png'],
                                              ['melf/melfe1.png', 'melf/melfe2.png'],
                                              ['melf/melfs1.png', 'melf/melfs2.png'],
                                              ['melf/melfw1.png', 'melf/melfw2.png'],
                                              4)
        sy, sx = 1, 1
        self.pentity = Entity(self.psprite, sy, sx)

        self.gsprite = EntitySprite.frompaths(screen,
                                              ['goblin/goblinn1.png', 'goblin/goblinn2.png'],
                                              ['goblin/gobline1.png', 'goblin/gobline2.png'],
                                              ['goblin/goblins1.png', 'goblin/goblins2.png'],
                                              ['goblin/goblinw1.png', 'goblin/goblinw2.png'],
                                              4)
        self.mkinds = {
            'GOBLIN': Goblin(self.gsprite),
        }

        self.pprofile = AnimatedSprite.scalefrompaths(screen,
                                                      ['melf/melfs1.png',
                                                       'melf/melfs2.png'],
                                                      scale=2)

        self.action = False

        self.lindex = 0
        self.zindex = 0
        self.zoff = 0
        self.zones = []
        lvl = self.create_top_level(sy, sx)
        self.zones.append(Zone(self, [lvl]))
        self.load_level(lvl.sy, lvl.sx)

        #self.monsters = self.curr_level().monsters
        self.update_text()
        self.update_stats()

    def update_text(self):
        self.nametxt = self.mktext(self.curr_player().name)
        self.titletxt = self.mktext("THE %s" % self.curr_player().title)
        if self.trading:
            self.zonetxt = self.mktext("TRADING")
        else:
            self.zonetxt = self.mktext("ZONE %d/%d" % (self.zindex + 1, self.lindex + 1))
    
    def update_stats(self):
        p = self.curr_player()
        self.hptxt = self.mktext("HP   %d/%d" % (p.hp, p.maxhp))
        self.goldtxt = self.mktext("GOLD %4d" % self.gold)

    def other_player(self):
        return self.zones[self.zindex + self.zoff].player

    def curr_player(self):
        return self.zones[self.zindex].player

    def curr_level(self):
        if self.trading:
            return self.tradelevel
        else:
            return self.zones[self.zindex].levels[self.lindex]

    def curr_zone(self):
        return self.zones[self.zindex]

    def tint_player_tile(self):
        #self.psprite.swaprgb(self.curr_player().locs)
        p = self.curr_player()
        p.entity.sprite.swaprgb(p.locs)
        self.pprofile.swaprgb(p.locs)

    def tint_level_tiles(self):
        #t = self.curr_level().tint
        t = self.curr_zone().tint
        tinter = lambda x: x ^ t
        self.walltile.tint(tinter)
        self.opentile.tint(tinter)
        self.doortile.tint(tinter)

    def create_trade(self):
        self.trading = True
        self.monsters = []
        self.maze = self.trademaze
        self.tiles = []

        self.tiles = []
        for y in range(0, self.theight):
            row = []
            for x in range(0, self.twidth):
                row.append(self.walltile)
            self.tiles.append(row)

        for y in range(1, self.theight - 1):
            for x in range(1, self.twidth - 1):
                self.tiles[y][x] = self.opentile

        self.tiles[10][5] = self.doortile
        self.maze.cave[10][5] = maze.TradeDoor()

        lvl = self.curr_level()
        lvl.items = {}

        op = self.other_player()
        for i in range(0, 3):
            if op.inventory[i] is not None:
                item = op.inventory[i]
                y, x = 6, 7 - i * 2
                lvl.items[(y, x)] = item

        self.pentity = self.curr_player().entity
        self.pentity.y = 9
        self.pentity.sprite.turn(0)
        self.pentity.x = 5

        self.update_text()
        self.update_stats()

    def create_maze(self, sy=0, sx=0, seed=999):
        self.maze = maze.Maze(self.theight, self.twidth, sy=sy, sx=sx,
                              nexits=1, seed=seed)
        nopen = 2 + self.lindex + random.randint(0, 8)
        self.tiles = []
        for y in range(0, len(self.maze.cave)):
            row = self.maze.cave[y]
            row2 = []
            for x in range(0, len(row)):
                cell = row[x]
                if cell.isstart:
                    if self.lindex > 0:
                        row2.append(self.starttile)
                    else:
                        row2.append(self.opentile)
                elif cell.isexit:
                    row2.append(self.exittile)
                elif cell.iswall:
                    row2.append(self.walltile)
                else:
                    row2.append(self.opentile)
            self.tiles.append(row2)

        if self.lindex == 0:
            ly, lx = self.theight - 2, self.twidth - 1
            if self.zindex > 0:
                self.tiles[1][0] = self.doortile
                self.maze.cave[1][0] = maze.LeftDoor()
            self.tiles[ly][lx] = self.doortile
            self.maze.cave[ly][lx] = maze.RightDoor()

        tiles = []
        for i in range(1, 11, 2):
            for j in range(1, 11, 2):
                if i != sy or j != sx:
                    tiles.append((i, j))
        random.shuffle(tiles)

        n = min(3 + (self.zindex / 2) + (self.lindex * 3) / 2, 30)
        self.monsters = []
        for y, x in tiles[:n]:
            speed = 4 + self.zindex
            self.monsters.append(self.mkinds['GOBLIN'].create(y, x, speed))

        self.pentity = self.curr_player().entity
        
        self.update_text()
        self.update_stats()
        self.started = False

    def adjacent(self, y, x):
        if y > 1: yield y - 1, x
        if y < self.theight - 1: yield y + 1, x
        if x > 1: yield y, x - 1
        if x < self.twidth - 1: yield y, x + 1
        raise StopIteration
    
    def draw(self, screen):
        # redraw the background
        screen.fill(self.bgcolor)

        # draw the maze
        for y in range(0, self.theight):
            for x in range(0, self.twidth):
                vx = self.xoff + x * 20
                vy = self.yoff + y * 20
                self.tiles[y][x].blitme((vx, vy))

        player = self.curr_player()
        lvl = self.curr_level()

        # draw items
        for y, x in lvl.items:
            vx, vy = self.xoff + x * 20, self.yoff + y * 20
            item = lvl.items[(y, x)]
            if item.isgold:
                self.goldtile.blitme((vx, vy))
            elif item.istime:
                self.timetile.blitme((vx, vy))
            elif item.ishealth:
                self.healthtile.blitme((vx, vy))
            else:
                item.kind.small.blitme((vx, vy))

        # draw the player
        self.pentity.sprite.update(pygame.time.get_ticks())
        gy, gx = self.pentity.getpos()
        self.pentity.sprite.blitme((self.xoff + gx, self.yoff + gy), player.isinverse())

        # draw an action?
        if self.action:
            item = player.inventory[player.equipped]
            y, x = self.pentity.y, self.pentity.x
            item.kind.render(self, item, y, x)
            item.kind.doeffect(self, item)

        self.monsters = [m for m in self.monsters if not m.isdead()]

        # draw monsters
        for monster in self.monsters:
            monster.update(self)
            e = monster.entity
            #monster.brain.doturn(self, monster)
            e.sprite.update(pygame.time.get_ticks())
            my, mx = e.getpos()
            e.sprite.blitme((self.xoff + mx, self.yoff + my))

        # draw something else
        maplimits = (40 + self.theight * 20, 40 + self.twidth * 20)
        ymargin = 20
        xmargin = 40 + self.twidth * 20

        screen.blit(self.zonetxt, (20, maplimits[0]))

        self.timetile.blitme((20, maplimits[0] + 40))
        screen.blit(self.timetxt, (40, maplimits[0] + 40))

        self.goldtile.blitme((20, maplimits[0] + 60))
        screen.blit(self.goldtxt, (40, maplimits[0] + 60))

        screen.blit(self.nametxt, (maplimits[1], 20))
        screen.blit(self.titletxt, (maplimits[1], 40))
        screen.blit(self.hptxt, (maplimits[1] + 80, 80))

        self.pprofile.update(pygame.time.get_ticks())
        self.pprofile.blitme((maplimits[1], 80))

        # draw inventory
        for i, item in enumerate(player.inventory):
            if item is not None:
                item.kind.big.blitme((maplimits[1] + 50 * i, 140))
            if i == player.equipped:
                self.selecttile.blitme((maplimits[1] + 50 * i + 10, 180))

        if self.trading:
            op = self.other_player()
            op.entity.sprite.update(pygame.time.get_ticks())

            op.entity.sprite.turn(2)
            oy, ox = 4 * 20, 5 * 20

            op.entity.sprite.blitme((self.xoff + ox, self.yoff + oy))

            self.firetile.blitme((self.xoff + 160, self.yoff + 100))
            self.firetile.blitme((self.xoff + 40, self.yoff + 100))

            screen.blit(self.dangertxt1, (maplimits[1] - 20, 260))
            screen.blit(self.dangertxt2, (maplimits[1] - 20, 280))
            screen.blit(self.dangertxt3, (maplimits[1] - 20, 300))


        # flip the screen
        pygame.display.flip()

    def create_top_level(self, sx, sy):
        return Level(self, None, random.randint(0, 1 << 24), sy, sx)

    def load_level(self, sy, sx):
        lvl = self.curr_level()
        self.create_maze(sy=sy, sx=sx, seed=lvl.seed)
        self.tint_player_tile()
        self.tint_level_tiles()

    def quit(self):
        self.done = True

    def toggle_fullscreen(self):
        pygame.display.toggle_fullscreen()

    def debug_levelgen(self):
        y, x = self.maze.exits[0]
        n = random.randint(0, 255)
        self.tint_level_tiles()
        self.create_maze(y, x)

    def prevtrade(self):
        if self.zindex == 0: return

        self.zoff = 1
        self.zindex -= 1

        lvl = self.curr_level()
        self.tint_player_tile()
        self.tint_level_tiles()
        self.create_trade()

    def nexttrade(self):
        self.zoff = -1
        if self.zindex == len(self.zones) - 1:
            lvl = self.create_top_level(1, 0)
            p = self.curr_player()
            self.zindex += 1
            self.zones.append(Zone(self, [lvl], forbid=p.locs))
        else:
            self.zindex += 1

        lvl = self.curr_level()
        self.tint_player_tile()
        self.tint_level_tiles()
        self.create_trade()

    def prevzone(self):
        self.trading = False


        lvl = self.curr_level()
        #self.tint_player_tile()
        #self.tint_level_tiles()
        self.pentity.y = self.theight - 2
        self.pentity.x = self.twidth - 1
        self.create_maze(lvl.sy, lvl.sx, lvl.seed)

    def nextzone(self):
        self.trading = False
        #if self.zindex == len(self.zones) - 1:
        #    lvl = self.create_top_level(1, 0)
        #    self.zones.append(Zone(self, [lvl], forbid=self.curr_zone().player.locs))
        #self.zindex += 1
        #
        lvl = self.curr_level()
        #self.tint_player_tile()
        #self.tint_level_tiles()
        self.pentity.y = 1
        self.pentity.x = 0
        self.create_maze(self.pentity.y, self.pentity.x, lvl.seed)

    def prevlevel(self):
        if self.lindex == 0: return
        self.lindex -= 1

        lvl = self.curr_level()
        self.tint_player_tile()
        self.tint_level_tiles()
        self.create_maze(lvl.sy, lvl.sx, lvl.seed)

    def nextlevel(self):
        levels = self.zones[self.zindex].levels
        if self.lindex == len(levels) - 1:
            lvl = self.curr_level()
            levels.append(Level(self, lvl.seed, lvl.next, self.pentity.y, self.pentity.x))
        self.lindex += 1

        lvl = self.curr_level()
        self.tint_player_tile()
        self.tint_level_tiles()
        self.create_maze(self.pentity.y, self.pentity.x, lvl.seed)

    def mktime(self, secs):
        return (secs / 60, secs % 60)

    def run(self):

        import time
        screen.blit(self.mktext("QUENGON: DUNGEON RELAY"), (20, 40))
        screen.blit(self.mktext("BY ERIK OSHEIM"), (80, 80))
        screen.blit(self.mktext("LUDUM DARE 20, 2011"), (40, 120))
        pygame.display.flip()
        time.sleep(10)

        #music = pygame.mixer.music.load('quengon.mp3')
        #pygame.mixer.music.play(-1)

        self.dirtytime = False

        self.t0 = time.time()
        self.secs = 0
        self.time = 120

        clock = pygame.time.Clock()

        self.dangertxt1 = self.mktext("IT'S DANGEROUS")
        self.dangertxt2 = self.mktext("TO GO ALONE...")
        self.dangertxt3 = self.mktext("TAKE THIS!")

        self.timetxt = self.mktext("TIME %02d:%02d" % self.mktime(self.time))

        self.done = False
        self.draw(screen)

        moves = []
        self.action = False

        # mighty main loop!
        while not self.done:
            # limit fps to 50
            time_passed = clock.tick(50)
            #time_passed = clock.tick(10)

            p = self.curr_player()
            if p.isdead():
                print 'player died'
                break

            secs = int(time.time() - self.t0)
            if self.dirtytime or secs > self.secs:
                self.dirtytime = False
                left = self.time - secs
                if left < 0:
                    print "out of time!!!!"
                    break

                self.secs = secs
                m = left / 60
                s = left % 60
                self.timetxt = self.mktext("TIME %02d:%02d" % (m, s))

            # stairs?
            if not self.started:
                pass
            elif not self.pentity.ismoving():
                sq = self.maze.cave[self.pentity.y][self.pentity.x]
                if sq.isexit:
                    self.nextlevel()
                elif sq.isstart:
                    self.prevlevel()
                elif sq.isleft:
                    self.prevtrade()
                elif sq.isright:
                    self.nexttrade()
                elif sq.istrade:
                    if self.zoff == -1:
                        self.nextzone()
                    else:
                        self.prevzone()
    
            # respond to events
            for event in pygame.event.get():
            
                if event.type == pygame.KEYUP:
                    if event.key == pygame.K_UP:
                        moves = [x for x in moves if x != 0]
                    elif event.key == pygame.K_RIGHT:
                        moves = [x for x in moves if x != 1]
                    elif event.key == pygame.K_DOWN:
                        moves = [x for x in moves if x != 2]
                    elif event.key == pygame.K_LEFT:
                        moves = [x for x in moves if x != 3]

                    elif self.action and self.action.wait == event.key:
                        item = p.inventory[p.equipped]
                        item.acting = False

                elif event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        self.quit()
                    elif event.key == pygame.K_x and event.mod == 64:
                        self.quit()
                    elif event.key == pygame.K_p and event.mod == 64:
                        y, x = self.pentity.y, self.pentity.x
                        print self.pentity.ismoving()
                        sq = self.maze.cave[y][x]
                        print sq
                        print sq.isexit
                        print self.pentity.yy, self.pentity.xx
                    elif event.key == pygame.K_r and event.mod == 64:
                        self.debug_levelgen()
                    elif event.key == pygame.K_TAB:
                        self.toggle_fullscreen()

                    elif event.key == pygame.K_UP:
                        moves.append(0)
                    elif event.key == pygame.K_RIGHT:
                        moves.append(1)
                    elif event.key == pygame.K_DOWN:
                        moves.append(2)
                    elif event.key == pygame.K_LEFT:
                        moves.append(3)

                    elif event.key == pygame.K_z:
                        if not self.action:
                            p.equipped = (p.equipped - 1) % 3
                    elif event.key == pygame.K_x:
                        if not self.action:
                            p.equipped = (p.equipped + 1) % 3

                    elif event.key == pygame.K_f:
                        self.pentity.startpush(3, 3, 20)

                    elif event.key == pygame.K_SPACE:
                        if self.action:
                            pass
                        else:
                            item = p.inventory[p.equipped]
                            if item is not None:
                                item.kind.act(self, item)
                                self.action = item.kind

                    else:
                        print event
                elif event.type == pygame.QUIT:
                    self.quit()

            if self.pentity.ispushed():
                if self.action:
                    self.action.cancel(item)
                    self.action = None
                self.pentity.keeppushing(self)
            elif self.action and self.action.pinned:
                pass
            elif self.pentity.ismoving():
                self.started = True
                self.pentity.keepmoving(self)
            elif moves:
                self.pentity.move(moves[0], self)

            p.update(self)

            if self.action:
                item = p.inventory[p.equipped]
                self.action = item.kind.update(self, item)

            for monster in self.monsters:
                if monster.entity.ispushed():
                    monster.entity.keeppushing(self)
                else:
                    monster.brain.doturn(self, monster)

            # detect monsters hitting player
            py1, px1, py2, px2 = self.pentity.getbox()
            for monster in self.monsters:
                if p.isimmune(): continue

                my1, mx1, my2, mx2 = monster.entity.getbox()
                if my1 >= py2 or my2 <= py1: continue
                if mx1 >= px2 or mx2 <= px1: continue

                ycmp = cmp(my1, py1)
                xcmp = cmp(mx1, px1)

                if my1 == py1:
                    d = 2 + cmp(mx1, px1)
                elif mx1 == px1:
                    d = 1 + cmp(py1, my1)
                else:
                    d = (self.pentity.sprite.facing + 2) % 4
                p.damage(self, d)
                monster.entity.startpush((monster.entity.sprite.facing + 2) % 4, 2, 10)
                self.update_stats()

            self.draw(screen)

        deadtxt = self.mktext("YOU HAVE DIED!!!")
        screen.blit(deadtxt, (20, 20))
        screen.blit(deadtxt, (20, 40))
        screen.blit(deadtxt, (20, 60))
        pygame.display.flip()
        time.sleep(2)
        #self.draw(screen)


class TradeMaze(object):
    def __init__(self, h, w):
        self.cave = []

        for y in range(0, h):
            row = []
            for x in range(0, w):
                row.append(maze.Wall())
            self.cave.append(row)

        for y in range(1, h - 1):
            for x in range(1, w - 1):
                self.cave[y][x] = maze.Open()

if __name__ == "__main__":
    width, height = 512, 384
    pygame.init()
    pygame.font.init()
    screen = pygame.display.set_mode((width, height), 0, 32)
    g = Game(screen)
    g.run()
    print 'quitting...'
    sys.exit()
