import math
import random
import sys

class Base(object):
    isstart = False
    iswall = False
    isexit = False
    isleft = False
    isright = False
    istrade = False

class Start(Base):
    isstart = True

class Wall(Base):
    iswall = True

class Open(Base):
    pass

class Exit(Base):
    isexit = True

class LeftDoor(Base):
    isleft = True

class RightDoor(Base):
    isright = True

class TradeDoor(Base):
    istrade = True

class Maze(object):
    def __init__(self, h, w, sy=0, sx=0, nexits=1, nopen=3, seed=0xdeadbeef):
        random.seed(seed)
        assert h % 2 == 1
        assert w % 2 == 1
        self.h = h
        self.w = w
        self.sy = sy
        self.sx = sx
        self.start = (self.sy, self.sx)
        self.exits = []

        self.nopen = nopen

        self.generate()
        self.place_start()
        self.save_distances()

        for i in range(0, nexits):
            self.place_exit()

        random.shuffle(self.maybe)
        for y, x in self.maybe[:self.nopen]:
            self.cave[y][x] = Open()

    def iterate(self):
        for y in range(0, self.h):
            for x in range(0, self.w):
                yield y, x
        raise StopIteration

    def open_iterate(self):
        for y, x in self.iterate():
            if not self.cave[y][x].iswall: yield y, x
        raise StopIteration

    def adjacent(self, y, x):
        if y > 0: yield y - 1, x
        if y < self.h - 1: yield y + 1, x
        if x > 0: yield y, x - 1
        if x < self.w - 1: yield y, x + 1
        raise StopIteration

    def open_directions(self, y, x):
        if y > 0 and not self.cave[y - 1][x].iswall: yield 0
        if x < self.w - 1 and not self.cave[y][x + 1].iswall: yield 1
        if y < self.h - 1 and not self.cave[y + 1][x].iswall: yield 2
        if x > 0 and not self.cave[y][x - 1].iswall: yield 3
        raise StopIteration

    def open_adjacent(self, y, x):
        for y, x in self.adjacent(y, x):
            if not self.cave[y][x].iswall:
                yield y, x
        raise StopIteration

    def place_start(self):
        #print 'place', self.sy, self.sx
        self.cave[self.sy][self.sx] = Start()

    def save_distances(self):
        self.distances = []
        for row in self.cave:
            self.distances.append([10000] * len(row))
        self.recalculate_distances()

    def recalculate_distances(self):
        queue = [(0, self.sy, self.sx)] + [(0, y, x) for y, x, in self.exits]
        seen = set()
        while queue:
            d, y, x = queue.pop(0)
            if self.distances[y][x] > d:
                self.distances[y][x] = d
            seen.add((y, x))

            for y2, x2 in self.open_adjacent(y, x):
                if (y2, x2) in seen:
                    continue
                queue.append((d + 1, y2, x2))

    def get_farthest_point(self):
        d = 0
        py = None
        px = None
        for y in range(0, self.h):
            for x in range(0, self.w):
                if self.cave[y][x].iswall:
                    pass
                elif y == self.h - 2 and x == self.w - 2:
                    pass
                elif len(list(self.open_adjacent(y, x))) > 1:
                    pass
                elif self.distances[y][x] > d:
                    d = self.distances[y][x]
                    py = y
                    px = x
        return py, px

    def place_exit(self):
        y, x = self.get_farthest_point()
        self.cave[y][x] = Exit()
        self.exits.append((y, x))
        self.recalculate_distances()

    def place_item(self):
        pass

    # Given an adjoining wall (a wall which separates two labyrinth cells)
    # set a and b to point to the cell indices which are separated. Used by
    # labyrinth_gen().
    def lab_get_adjoin(self, y, x):
        if x % 2 == 1: # xyz
            p1 = (y - 1, x)
            p2 = (y + 1, x)
        else:
            p1 = (y, x - 1)
            p2 = (y, x + 1)
        return (p1, p2)
    
    def is_start_connected(self):
        for p in self.open_adjacent(self.sy, self.sx):
            return True
        return False

    # Build a labyrinth level.
    def generate(self):
        h = self.h
        w = self.w

        # This will be a two-dimensional list of our cells.
        self.cave = cave = []

        # This will be a map giving each unconnected cell "area" its own color.
        sets = {}

        # This is a list of all the tile coords, which we will later randomize.
        walls = []
    
        # Allocate all the grids, start them off as walls.
        for y in range(0, h):
            row = []
            for x in range(0, w):
                row.append(Wall())
                walls.append((y, x))
                sets[(y, x)] = y + x * w
            cave.append(row)

        # Make the areas where x and y are both odd open. # xyz
        for y in range(1, h, 2):
            for x in range(1, w, 2):
                cave[y][x] = Open()
    
        # Randomize the wall list.
        random.shuffle(walls)
    
        # For each adjoining wall, look at the cells it divides. If they aren't
        # in the same set, remove the wall and join their sets.
        #
        # This is a randomized version of Kruskal's algorithm.
        self.maybe = []
        for wall in walls:
            y, x = wall
    
            # If this cell isn't an adjoining wall, skip it */
            if x < 1 or x > w - 2: continue
            if y < 1 or y > h - 2: continue
            if (x < 2 and y < 2) or (x > w - 3 and y > h - 3): continue
            if (x % 2 == y % 2): continue
        
            # Figure out which cells are separated by this wall */
            p1, p2 = self.lab_get_adjoin(y, x)

            # We want the "start" position to be a dead-end
            if p1 == self.start or p2 == self.start:
                if self.is_start_connected(): continue

            # If the cells aren't connected, kill the wall and join the sets */
            #continue
            if sets[p1] != sets[p2]:
                cave[y][x] = Open()

                # Now change all the cells of p2's color to p1's color.
                replace = sets[p2]
                for y2 in range(0, h):
                    for x2 in range(0, w):
                        if sets[(y2, x2)] == replace:
                            sets[(y2, x2)] = sets[p1]
            else:
                self.maybe.append((y, x))

if __name__ == "__main__":
    m = Maze()
    print m
